// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
// 
// Licensed under the MIT license <LICENSE-MIT or http:// opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.

#include "pvi.h"

template <typename T> void pvi::free_vec( T & t ) {
    T tmp;
    t.swap( tmp );
}

// the point q has the following components:
// {q.q1,q.q2,q.q3,q.qinf,q.q21,q.q31,q.q32,o1,o2,o3,o4,orbit's #,size}
// the first 11 components correspond to:
// (q1,q2,q3,qinf,q21,q31,q32,\omega_1,\omega_2,\omega_3,\omega_4)

// Save results on file
void pvi::save_E45_on_file( std::vector<pvi::pvi_coords> vec , std::string str ){
     std::ofstream out_file;
     try{
        out_file.open( str.c_str() , std::ios::binary );
        for( auto& x : vec ){
            pvi::pvi_coords y = pvi::pvi_coords();
            y.q1 = x.q1;
            y.q2 = x.q2;
            y.q3 = x.q3;
           	y.qinf = x.qinf;
            y.q21 = x.q21;
            y.q31 = x.q31;
            y.q32 = x.q32;
            y.o1 = x.o1;
            y.o2 = x.o2;
            y.o3 = x.o3;
            y.o4 = x.o4;
            y.num_orb = x.num_orb;
            y.num_points = x.num_points;
            out_file.write(reinterpret_cast<char *>(&y), sizeof(y));
		}
		out_file.close();
    }catch (std::ofstream::failure e) {
        std::cerr << "Error: ";
    }
}
// Load results from file
void pvi::load_E45_from_file( std::vector<pvi::pvi_coords>& vec , std::string str ){
    // container must be empty
    vec.clear();
    pvi::pvi_coords q = pvi::pvi_coords();
    try{
        std::ifstream in_file( str.c_str() , std::ios::binary );
		pvi::pvi_coords y = pvi::pvi_coords();
        while( !in_file.eof() ){
//              std::cout << "Err" << std::endl;
            in_file.read(reinterpret_cast<char *>(&y), sizeof(y));
            q.q1 = y.q1;
            q.q2 = y.q2;
            q.q3 = y.q3;
           	q.qinf = y.qinf;
            q.q21 = y.q21;
            q.q31 = y.q31;
            q.q32 = y.q32;
            q.o1 = y.o1;
            q.o2 = y.o2;
            q.o3 = y.o3;
            q.o4 = y.o4;
            q.num_orb = y.num_orb;
            q.num_points = y.num_points;
            vec.push_back( q );
        }
    }catch (std::ifstream::failure e) {
        std::cerr << "Error: ";
    }
}

// print coordinates on screen
void pvi::print_q_point( pvi::pvi_coords q ){
    std::cout << '{' << q.q1 << " , " << q.q2 << " , " << q.q3 << " , " << q.qinf << " , " << q.q21 << " , " << q.q31 << " , " << q.q32 << " , " << q.o1 << " , " << q.o2 << " , " << q.o3 << " , " << q.o4 << '}' << std::endl;    
};

//  avoid propagation of errors
//  checking every components of q
void pvi::correct_values( pvi::pvi_coords& q ){
	q.q1 = gen_value(q.q1);
	q.q2 = gen_value(q.q2);
	q.q3 = gen_value(q.q3);
	q.qinf = gen_value(q.qinf);
	q.q21 = gen_value(q.q21);
	q.q31 = gen_value(q.q31);
	q.q32 = gen_value(q.q32);
	q.o1 = gen_value(q.o1);
	q.o2 = gen_value(q.o2);
	q.o3 = gen_value(q.o3);
	q.o4 = gen_value(q.o4);
};

// generate a key for the vector q
std::size_t pvi::make_hash_q( pvi::pvi_coords q ){
    std::ostringstream s;
    s.str(" ");
    s << q.q1 << ' ';
    s << q.q2 << ' ';
    s << q.q3 << ' ';
    s << q.qinf << ' ';
    s << q.q21 << ' ';
    s << q.q31 << ' ';
    s << q.q32 << ' ';
    s << q.o1 << ' ';
    s << q.o2 << ' ';
    s << q.o3 << ' ';
    s << q.o4 << ' ';
    return std::hash<std::string>()(s.str());
}
// insert a new vector q in map 
// return ret : ret.first is an iterator pointing to either the newly inserted element or to the element with an equivalent key in the map
// 			   ret.second is true if a new element was inserted or false if the element already existed
void pvi::put_in_map_q( pvi::pvi_coords q , std::map<size_t, pvi::pvi_coords >& map , std::pair<std::map<size_t, pvi::pvi_coords >::iterator,bool>& ret){
    ret = map.insert(std::pair<size_t, pvi::pvi_coords >(make_hash_q( q ),q));    
}

// Jimbo-Fricke cubic relation
double pvi::jf_cubic( pvi::pvi_coords q ){ 
    double result =   -4.0 + pow(q.q1,2.0) + pow(q.q2,2.0) - 
        q.q1*q.q2*q.q21 + pow(q.q21,2.0) + pow(q.q3,2.0) - q.q1*q.q3*q.q31 + pow(q.q31,2.0) - q.q2*q.q3*q.q32 + 
        q.q21*q.q31*q.q32 + pow(q.q32,2.0) + q.q1*q.q2*q.q3*q.qinf - q.q21*q.q3*q.qinf - q.q2*q.q31*q.qinf - 
        q.q1*q.q32*q.qinf + pow(q.qinf,2.0);
    return result; 
}

// Generators of the pure braid group P_3
pvi::pvi_coords pvi::pb1( pvi::pvi_coords q ){
	pvi::pvi_coords result; 
	result.q1 = q.q1; result.q2 = q.q2; result.q3 = q.q3; result.qinf = q.qinf; result.q21 = q.q21; result.q31 = q.o2 - q.q31 - q.q21*q.q32; result.q32 = q.o1 - q.q32 - q.q21*(q.o2 - q.q31 - q.q21*q.q32); result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
     correct_values( result );
     return result; 
};
pvi::pvi_coords pvi::pb2( pvi::pvi_coords q ){
	pvi::pvi_coords result; 
	result.q1 = q.q1; result.q2 = q.q2; result.q3 = q.q3; result.qinf = q.qinf; result.q21 = q.o3 - q.q21 - q.q31*q.q32; result.q31 = q.o2 - q.q31 - q.q32*(q.o3 - q.q21 - q.q31*q.q32); result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
     correct_values( result );
     return result; 
};
pvi::pvi_coords pvi::pb3( pvi::pvi_coords q ){
	pvi::pvi_coords result; 
	{result.q1 = q.q1; result.q2 = q.q2; result.q3 = q.q3; result.qinf = q.qinf; result.q21 = q.o3 - q.q21 - q.q31*(q.o1 - q.q21*q.q31 - q.q32); result.q31 = q.q31; result.q32 = q.o1 - q.q21*q.q31 - q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;}
     correct_values( result );
     return result; 
};
// recursive function that generates the orbit of q under the action of the pure braid group P3
void pvi::gen_orbit_( pvi::pvi_coords q , std::vector<pvi::pvi_coords >& orbit_points ){
		// set a maximum otherwise the function could run forever
		max_recursion--;
		// is the maximum is reached then quit
		if(max_recursion < 0) return;
		// create and save the unique key associated to q
		put_in_map_q( q , hash_table , ret );
		// if the key already exists then check component by component if q and the element q' := ret.first->second are the same vector :
		// q.qi = q'.qi
		// q.qij = q'.qij
		if( ret.second == false ){
			// if they are not the same vector then trigger an error
			if(										
				!(fabs(ret.first->second.q1 - q.q1) < SMALL &&
				fabs(ret.first->second.q2 - q.q2) < SMALL &&
				fabs(ret.first->second.q3 - q.q3) < SMALL &&
				fabs(ret.first->second.qinf - q.qinf) < SMALL &&
				fabs(ret.first->second.q21 - q.q21) < SMALL &&
				fabs(ret.first->second.q31 - q.q31) < SMALL &&
				fabs(ret.first->second.q32 - q.q32) < SMALL)
			){
				std::cout << "Err. elements not equal" << std::endl;
				print_q_point(ret.first->second);
				print_q_point(q);
				return;
			}
			return;
		}else{
			// if the key doesn't exist in the table, insert the element in the orbit orbit_points
			orbit_points.push_back( q ); 
		}
		// since q is a new element in the orbit, apply to it all the three generators, pbi, of the pure braid group P3
		pvi::pvi_coords tpb1 = pb1( q );
		pvi::pvi_coords tpb2 = pb2( q );
		pvi::pvi_coords tpb3 = pb3( q );
		// call the function for every element q' = pbi( q )
		gen_orbit_( tpb1 , orbit_points );
		gen_orbit_( tpb2 , orbit_points );
		gen_orbit_( tpb3 , orbit_points );
}

//  24 solutions (q1,q2,q3,qinf) of system (o1,o2,o3,o4) 
pvi::pvi_coords pvi::d4_1( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =q.q1;

	result.q2 =q.q2;

	result.q3 =q.q3;

	result.qinf =q.qinf;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_2( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =-q.q1;

	result.q2 =-q.q2;

	result.q3 =-q.q3;

	result.qinf =-q.qinf;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_3( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =q.q2;

	result.q2 =q.q1;

	result.q3 =q.qinf;

	result.qinf =q.q3;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_4( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =q.q3;

	result.q2 =q.qinf;

	result.q3 =q.q1;

	result.qinf =q.q2;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_5( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =q.qinf;

	result.q2 =q.q3;

	result.q3 =q.q2;

	result.qinf =q.q1;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_6( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =-q.q2;

	result.q2 =-q.q1;

	result.q3 =-q.qinf;

	result.qinf =-q.q3;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_7( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =-q.q3;

	result.q2 =-q.qinf;

	result.q3 =-q.q1;

	result.qinf =-q.q2;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_8( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =-q.qinf;

	result.q2 =-q.q3;

	result.q3 =-q.q2;

	result.qinf =-q.q1;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_9( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;
    
    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_10( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_11( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_12( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_13( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_14( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_15( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_16( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_17( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_18( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_19( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_20( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_21( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_22( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_23( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::d4_24( pvi::pvi_coords q ){
    pvi::pvi_coords result;

	result.q1 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q2 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.q3 =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

	result.qinf =(-(sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf)) + sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 - q.qinf) + sqrt(2.0 + q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 - q.q3)*sqrt(2.0 + q.qinf) + sqrt(2.0 - q.q1)*sqrt(2.0 - q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf) - sqrt(2.0 + q.q1)*sqrt(2.0 + q.q2)*sqrt(2.0 + q.q3)*sqrt(2.0 + q.qinf))/8.0;

    result.q21 = q.q21; result.q31 = q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points;
    correct_values( result );
    return result;
}
std::vector<pvi::pvi_coords > pvi::gen_24sol_qis_( pvi::pvi_coords q ){
    std::vector<pvi::pvi_coords > result = {
        d4_1(q),d4_2(q),d4_3(q),d4_4(q),d4_5(q),d4_6(q),d4_7(q),d4_8(q),d4_9(q),d4_10(q),d4_11(q),d4_12(q),d4_13(q),d4_14(q),d4_15(q),d4_16(q),d4_17(q),d4_18(q),d4_19(q),d4_20(q),d4_21(q),d4_22(q),d4_23(q),d4_24(q)
	};
	return result;
}

//  Generators group <sign_1,sign_2,sign_3>
pvi::pvi_coords pvi::sign1_( pvi::pvi_coords q ){
    pvi::pvi_coords result;
    result.q1 = -q.q1; result.q2 = q.q2; result.q3 = q.q3; result.qinf = -q.qinf; result.q21 = -q.q21; result.q31 = -q.q31; result.q32 = q.q32; result.o1 = q.o1; result.o2 = -q.o2; result.o3 = -q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::sign2_( pvi::pvi_coords q ){
    pvi::pvi_coords result;
    result.q1 = q.q1; result.q2 = -q.q2; result.q3 = q.q3; result.qinf = -q.qinf; result.q21 = -q.q21; result.q31 = q.q31; result.q32 = -q.q32; result.o1 = -q.o1; result.o2 = q.o2; result.o3 = -q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::sign3_( pvi::pvi_coords q ){
    pvi::pvi_coords result;
    result.q1 = q.q1; result.q2 = q.q2; result.q3 = -q.q3; result.qinf = -q.qinf; result.q21 = q.q21; result.q31 = -q.q31; result.q32 = -q.q32; result.o1 = -q.o1; result.o2 = -q.o2; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
// function that generates all the elments of the group <sign_1,sign_2,sign_3>
std::vector<pvi::pvi_coords > pvi::gen_sign_flips_( pvi::pvi_coords q ){
    std::vector<pvi::pvi_coords > result = {
        q,							// id
        sign1_( q ),				
		sign2_( q ),				
		sign3_( q ),
		sign2_( sign1_( q ) ),
		sign3_( sign1_( q ) ),
		sign3_( sign2_( q ) ),
		sign3_( sign2_( sign1_( q ) ) )
		}; 
        return result;
}

//  Generators group S3 = <(12),(123)>
pvi::pvi_coords pvi::perm123_( pvi_coords q ){
    pvi::pvi_coords result;
   	result.q1 = q.q3; result.q2 = q.q1; result.q3 = q.q2; result.qinf = q.qinf; result.q21 = q.q31; result.q31 = q.q32; result.q32 = q.q21; result.o1 = q.o3; result.o2 = q.o1; result.o3 = q.o2; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::perm12_( pvi_coords q ){
    pvi::pvi_coords result;
   	result.q1 = q.q2; result.q2 = q.q1; result.q3 = q.q3; result.qinf = q.qinf; result.q21 = q.q21; result.q31 = q.q32; result.q32 = q.q31; result.o1 = q.o2; result.o2 = q.o1; result.o3 = q.o3; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
// function that generates all the elments of the group  S3
std::vector<pvi::pvi_coords > pvi::gen_S3_permutations_( pvi::pvi_coords q ){
    std::vector<pvi::pvi_coords > result = {
        q,							// id
        perm12_( q ),				// (12)
		perm123_( q ),				// (123)
		perm12_( perm123_( q ) ),	// (23)
		perm123_( perm12_( q ) ),	// (13)
		perm123_( perm123_( q ) )	// (132)
		};
        return result;
}

//  Generators group <P13,P23>
pvi::pvi_coords pvi::P13_( pvi::pvi_coords q ){
    pvi::pvi_coords result;
    result.q1 = q.q3; result.q2 = q.q2; result.q3 = q.q1; result.qinf = q.qinf; result.q21 = q.q32; result.q31 = q.o2 - q.q31 - q.q21*q.q32; result.q32 = q.q21; result.o1 = q.o3; result.o2 = q.o2; result.o3 = q.o1; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
pvi::pvi_coords pvi::P23_( pvi::pvi_coords q ){
    pvi::pvi_coords result;
    result.q1 = q.q1; result.q2 = q.q3; result.q3 = q.q2; result.qinf = q.qinf; result.q21 = q.o2 - q.q31 - q.q21*q.q32; result.q31 = q.q21; result.q32 = q.q32; result.o1 = q.o1; result.o2 = q.o3; result.o3 = q.o2; result.o4 = q.o4; result.num_orb = q.num_orb; result.num_points = q.num_points; 
    correct_values( result );
    return result;
}
// recursive function that generates the orbit of q under the action of <P13,P23>
// (see comments on the function gen_orbit_)
void pvi::gen_P13_P23_group_( pvi::pvi_coords q , std::vector<pvi::pvi_coords >& orbit_points ){
		max_recursion--;
		if(max_recursion < 0) return;
		put_in_map_q( q , hash_table_p13p23 , ret );
		if( ret.second == false ){return;}else{ orbit_points.push_back( q ); }
		pvi::pvi_coords tp13 = P13_( q );
		pvi::pvi_coords tp23 = P23_( q );
		gen_P13_P23_group_( tp13 , orbit_points );
		gen_P13_P23_group_( tp23 , orbit_points );
}
pvi::pvi(){}
pvi:: ~pvi(){
    free_vec(E45);
//      std::cout << "Destructor Called. [pvi]" << std::endl;
}