// Copyright (c) 2013 - 2017 P. Calligaris and M. Mazzocco. 
// All rights reserved.
// 
// Licensed under the MIT license <LICENSE-MIT or http:// opensource.org/licenses/MIT>, 
// at your option. This file may not be copied, modified, or distributed except according
// to those terms.
//
// Algorithm to calculate finite monodromy groups

#include "pvi.h"
#include "g2.h"
#include "custom_library.h"
#include <complex.h>
#include <cmath>


std::complex<double> correct_value( std::complex<double> x ){
    double a,b;
    double my_small = 0.0000001;
    a = x.real();
    b = x.imag();
    if( fabs(x.real()) < my_small ) a = 0.0;
    if( fabs(x.imag()) < my_small ) b = 0.0;
    std::complex<double> result(a,b);
    return result;
}

bool check_is_in( std::vector< std::complex<double> > matrix, std::vector< std::vector< std::complex<double> > > orbit  ){
    double my_small = 0.0000001;
    for(int ii = 0; ii < orbit.size(); ii++){
        if( fabs( matrix.at(0).real() - orbit.at(ii).at(0).real()) < my_small && fabs( matrix.at(0).imag() - orbit.at(ii).at(0).imag()) < my_small &&
            fabs( matrix.at(1).real() - orbit.at(ii).at(1).real()) < my_small && fabs( matrix.at(1).imag() - orbit.at(ii).at(1).imag()) < my_small &&
            fabs( matrix.at(2).real() - orbit.at(ii).at(2).real()) < my_small && fabs( matrix.at(2).imag() - orbit.at(ii).at(2).imag()) < my_small &&
            fabs( matrix.at(3).real() - orbit.at(ii).at(3).real()) < my_small && fabs( matrix.at(3).imag() - orbit.at(ii).at(3).imag()) < my_small
         ){
            return true;
        }else if( fabs( matrix.at(0).real() + orbit.at(ii).at(0).real()) < my_small && fabs( matrix.at(0).imag() + orbit.at(ii).at(0).imag()) < my_small &&
            fabs( matrix.at(1).real() + orbit.at(ii).at(1).real()) < my_small && fabs( matrix.at(1).imag() + orbit.at(ii).at(1).imag()) < my_small &&
            fabs( matrix.at(2).real() + orbit.at(ii).at(2).real()) < my_small && fabs( matrix.at(2).imag() + orbit.at(ii).at(2).imag()) < my_small &&
            fabs( matrix.at(3).real() + orbit.at(ii).at(3).real()) < my_small && fabs( matrix.at(3).imag() + orbit.at(ii).at(3).imag()) < my_small
         ){
            return true;
        }
    }
    return false;
} 

bool is_equal(std::vector< std::complex<double> > matrix_a, std::vector< std::complex<double> > matrix_b ){
    double my_small = 0.0000001;
    return (fabs( matrix_a.at(0).real() - matrix_b.at(0).real()) < my_small && fabs( matrix_a.at(0).imag() - matrix_b.at(0).imag()) < my_small &&
            fabs( matrix_a.at(1).real() - matrix_b.at(1).real()) < my_small && fabs( matrix_a.at(1).imag() - matrix_b.at(1).imag()) < my_small &&
            fabs( matrix_a.at(2).real() - matrix_b.at(2).real()) < my_small && fabs( matrix_a.at(2).imag() - matrix_b.at(2).imag()) < my_small &&
            fabs( matrix_a.at(3).real() - matrix_b.at(3).real()) < my_small && fabs( matrix_a.at(3).imag() - matrix_b.at(3).imag()) < my_small);
    
}

std::vector< std::complex<double> > matrix_multiplication( std::vector< std::complex<double> > m1, std::vector< std::complex<double> > m2 ){
    return { correct_value(m1.at(0) * m2.at(0) + m1.at(1) * m2.at(2)), correct_value(m1.at(0) * m2.at(1) + m1.at(1) * m2.at(3)), 
                correct_value(m1.at(2) * m2.at(0) + m1.at(3) * m2.at(2)), correct_value(m1.at(2) * m2.at(1) + m1.at(3) * m2.at(3)) };
}

int main(int argc, char* argv[])
{
    std::cout << "Generate matrices M1,...,M4 for the 54 orbits" << std::endl;
    
	if (argc < 2) {
        std::cout << "Usage: " << argv[0] << " C2" << std::endl;
        return EXIT_FAILURE;
    }  
 
 	std::string file_name(argv[1]);
 
    g2 g;
	std::map<size_t, std::vector<double> > hash_table_loc,hash_table_loc_t;
	std::map<size_t, std::vector<double> >::iterator it;
	std::pair<std::map<size_t, std::vector<double> >::iterator,bool> ret;
	std::vector<double> double_p, double_p_t;
	g2::g2_coords p, permuted_p, braided_p;
    std::vector<g2::g2_coords> points_p,points_p_t;
	int sz_1 = 0;

    std::vector< std::vector< std::vector< std::complex<double> > > > classification = { { {0, -1.0, 1.0, 0}, {std::complex<double>(0,-1.0)/sqrt(2.0), -(1.0/sqrt(2.0)), 1.0/sqrt(2.0), std::complex<double>(0,1.0)/sqrt(2.0)}, {std::complex<double>(-0.5,-0.5), std::complex<double>(-0.5,0.5), std::complex<double>(0.5,0.5), std::complex<double>(-0.5,0.5)}, {0, std::complex<double>(0,1.0), std::complex<double>(0,1.0), 0} },  
    /*#2*/                                                                              { {0, -1.0, 1.0, 0}, {0, std::complex<double>(0,1.0), std::complex<double>(0,1.0), 0}, {0, 1.0, -1.0, 0}, {std::complex<double>(0.5,0.5), std::complex<double>(-0.5,0.5), std::complex<double>(0.5,0.5), std::complex<double>(0.5,-0.5)} },  
    /*#3*/                                                                              { {std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(2.0)), -0.25, 1.0, -0.5- std::complex<double>(0,1.0)/sqrt(2.0)}, {0.5+ std::complex<double>(0,1.0)/sqrt(2.0), 0.25, -1.0, 0.5- std::complex<double>(0,1.0)/sqrt(2.0)}, {std::complex<double>(0,0.5) + 1.0/sqrt(2.0), std::complex<double>(0,-0.25), std::complex<double>(0,-1.0), std::complex<double>(0,-0.5) + 1.0/sqrt(2.0)}, {0.5- std::complex<double>(0,1.0)/sqrt(2.0), 0.25, -1.0, 0.5+ std::complex<double>(0,1.0)/sqrt(2.0)} },  
    /*#4*/                                                                              { {(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), -0.3333333333333333, 1.0, (std::complex<double>(0,-3.0) - sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), (1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6., std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0)), (std::complex<double>(0,-3.0) - sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,-3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), -0.3333333333333333, 1.0, (std::complex<double>(0,3.0) + sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)), ((1.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/4.0, ((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))} },  
                                                                                        { {std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) - sqrt((5.0 + sqrt(5))/10.) + sqrt((5.0 + sqrt(5))/2.0)), (-5.0 + sqrt(5))/10., 1.0, std::complex<double>(0,0.025)*(std::complex<double>(0,10)*(1.0 + sqrt(5)) - 5*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5))))}, {(std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), -2.0/(5.0 + sqrt(5)), 1.0, (std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), -(1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (2.0 - 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,3.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {0.25 - sqrt(5)/4.0 - std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5)), (-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), -(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-3.0 + sqrt(5))), 0.25 - sqrt(5)/4.0 + std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5))} },  
                                                                                        { {std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5)) + (-1.0 + sqrt(5))/4.0, (-5.0 - sqrt(5))/10., 1.0, std::complex<double>(0,-1.0)/sqrt(10 - 2.0*sqrt(5)) + (-1.0 + sqrt(5))/4.0}, {(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0, 0, 0, (-1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0}, {(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0, 0, 0, (-1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0}, {0.25 - sqrt(5)/4.0 - std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5)), (1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/4.0, 0.25 - sqrt(5)/4.0 + std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5))} },  
                                                                                        { {(sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(-1.0 + sqrt(5)))/(4.0*sqrt(10 - 2.0*sqrt(5))), (-5.0 + sqrt(5))/10., 1.0, (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/10.) - std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0))/4.0}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (-1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/4.0, ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), -(-2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(5.0 + sqrt(5))), (8 - 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), (-2.0 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,3.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (2.0 - 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,3.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))} },  
                                                                                        { {1.0/sqrt(2.0), -0.5, 1.0, 1.0/sqrt(2.0)}, {0, 1.0/sqrt(2.0), -sqrt(2.0), 0}, {0, std::complex<double>(-0.5,-0.5), std::complex<double>(1.0,-1.0), 0}, {std::complex<double>(0,-1.0)/sqrt(2.0), std::complex<double>(0,0.5), std::complex<double>(0,1.0), std::complex<double>(0,1.0)/sqrt(2.0)} },  
                                                                                        { {0, -1.0, 1.0, 0}, {std::complex<double>(0,-1.0)/sqrt(3.0), 1.0/sqrt(2.0) - std::complex<double>(0,1.0)/sqrt(6), -(1.0/sqrt(2.0)) - std::complex<double>(0,1.0)/sqrt(6), std::complex<double>(0,1.0)/sqrt(3.0)}, {std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0)), -(1.0/sqrt(2.0)) - std::complex<double>(0,1.0)/sqrt(6), 1.0/sqrt(2.0) - std::complex<double>(0,1.0)/sqrt(6), std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0))}, {std::complex<double>(0,-1.0)*sqrt(0.6666666666666666), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0)), (3.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6., std::complex<double>(0,1.0)*sqrt(0.6666666666666666)} },  
                                                                                        { {-((std::complex<double>(0,-1.0) + sqrt(3.0))/sqrt(6)), -0.3333333333333333, 1.0, -((std::complex<double>(0,1.0) + sqrt(3.0))/sqrt(6))}, {0, std::complex<double>(0,-1.0)/sqrt(3.0), std::complex<double>(0,-1.0)*sqrt(3.0), 0}, {0, std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0)), (3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/2.0, 0}, {std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0)), (1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/(3.0*sqrt(2.0)), (-1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/sqrt(2.0), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0))} },  
                                                                                        { {((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 - sqrt(5))/6., 1.0, ((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {-(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/12., (std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), (std::complex<double>(0,3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) - sqrt(15))/(4.0*sqrt(3.0))}, {std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(15)), (1.0 + sqrt(5))/6., (1.0 - sqrt(5))/2.0, std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(15))}, {((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,1.0) + sqrt(3.0))*(3.0 + sqrt(5)), (1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/2.0, ((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))} },  
                                                                                        { {(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/10.) - std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0))/4.0, (-5.0 + sqrt(5))/10., 1.0, (sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(-1.0 + sqrt(5)))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {(sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(-1.0 + sqrt(5)))/(4.0*sqrt(10 - 2.0*sqrt(5))), (8 - 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (-8 + 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/10.) - std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0))/4.0}, {(std::complex<double>(0,0.5)*(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-8 + 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (8 - 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), (std::complex<double>(0,-0.5)*(1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) - sqrt((5.0 + sqrt(5))/10.) + sqrt((5.0 + sqrt(5))/2.0)), (8 - 4.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (-8 + 4.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), std::complex<double>(0,0.025)*(std::complex<double>(0,10)*(1.0 + sqrt(5)) - 5*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5))))} },  
                                                                                        { {std::complex<double>(0,1.0)/sqrt(3.0), -0.6666666666666666, 1.0, std::complex<double>(0,-1.0)/sqrt(3.0)}, {0, std::complex<double>(0,-1.0)*sqrt(0.6666666666666666), std::complex<double>(0,-1.0)*sqrt(1.5), 0}, {0, -(1.0/sqrt(2.0)) - std::complex<double>(0,1.0)/sqrt(6), (3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/(2.0*sqrt(2.0)), 0}, {std::complex<double>(0,-1.0)*sqrt(0.6666666666666666), (1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/(3.0*sqrt(2.0)), (std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0)))/sqrt(2.0), std::complex<double>(0,1.0)*sqrt(0.6666666666666666)} },  
                                                                                        { {(std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {(std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), -((1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/(4.0*(5.0 + sqrt(5))), ((1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/8., (std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), ((1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/(4.0*(5.0 + sqrt(5))), -((1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/8., ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) - sqrt((5.0 + sqrt(5))/10.) + sqrt((5.0 + sqrt(5))/2.0)), (5.0 - sqrt(5))/10., -1.0, std::complex<double>(0,0.025)*(std::complex<double>(0,10)*(1.0 + sqrt(5)) - 5*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5))))} },  
                                                                                        { {(1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0, (-5.0 - sqrt(5))/10., 1.0, (1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0}, {(std::complex<double>(0,-0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/(2.0*(-5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) - std::complex<double>(0,3.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), (std::complex<double>(0,0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))}, {-((-1.0 + sqrt(5))*(std::complex<double>(0,1.0)*(3.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5)))))/(4.0*sqrt(2.0*(5.0 + sqrt(5)))), 2.0/(-5.0 + sqrt(5)), 1.0, (std::complex<double>(0,0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))}, {(std::complex<double>(0,0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,-0.25)*(std::complex<double>(0,-4.0)*(2.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(5.0 + sqrt(5)), (8 + 4.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), (std::complex<double>(0,-0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {std::complex<double>(0,-0.5), -0.75, 1.0, std::complex<double>(0,0.5)}, {std::complex<double>(0,0.5), 0.75, -1.0, std::complex<double>(0,-0.5)}, {0.5, std::complex<double>(0,-0.75), std::complex<double>(0,-1.0), 0.5}, {std::complex<double>(0,-1.0)/sqrt(2.0), (std::complex<double>(0,-2.0) - sqrt(2.0))/4.0, (std::complex<double>(0,-2.0) + sqrt(2.0))/3.0, std::complex<double>(0,1.0)/sqrt(2.0)} },  
                                                                                        { {(std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {0.5+ (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), -(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/4.0, 0.5- (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/10.) - std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0))/4.0, (-2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(5.0 + sqrt(5))), (2.0 - 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/8., (sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(-1.0 + sqrt(5)))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {(std::complex<double>(0,-0.5)*(1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (-1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/4.0, (std::complex<double>(0,0.5)*(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))} },  
                                                                                        { {0.5+ (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (-5.0 - sqrt(5))/10., 1.0, 0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {(1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0, 0, 0, (1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0}, {(1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0, 0, 0, (1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0}, {(std::complex<double>(0,0.5)*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), -(2.0 - 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (-2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/8., -0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))} },  
                                                                                        { {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), (-5.0 + sqrt(5))/10., 1.0, ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))}, {0.5- (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (-1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/4.0, 0.5+ (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {std::complex<double>(0,0.025)*(std::complex<double>(0,10)*(1.0 + sqrt(5)) - 5*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), 2.0/(5.0 + sqrt(5)), -1.0, std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) - sqrt((5.0 + sqrt(5))/10.) + sqrt((5.0 + sqrt(5))/2.0))}, {std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) - sqrt((5.0 + sqrt(5))/10.) + sqrt((5.0 + sqrt(5))/2.0)), (-5.0 + sqrt(5))/10., 1.0, std::complex<double>(0,0.025)*(std::complex<double>(0,10)*(1.0 + sqrt(5)) - 5*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5))))} },  
                                                                                        { {(3.0 - std::complex<double>(0,1.0)*sqrt(15))/6., -0.3333333333333333, 1.0, (3.0 + std::complex<double>(0,1.0)*sqrt(15))/6.}, {((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), ((1.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/12., ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(-1.0 + sqrt(5)))/(2.0*(-3.0 + sqrt(5))), ((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {-((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)), (std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), -((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)), (std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), ((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))} },  
                                                                                        { {std::complex<double>(0.5,-0.5), -0.5, 1.0, std::complex<double>(0.5,0.5)}, {std::complex<double>(0,-1.0)/sqrt(2.0), std::complex<double>(0.5,-0.5)/sqrt(2.0), std::complex<double>(-1.0,-1.0)/sqrt(2.0), std::complex<double>(0,1.0)/sqrt(2.0)}, {std::complex<double>(-0.5,0.5), -0.5, 1.0, std::complex<double>(-0.5,-0.5)}, {std::complex<double>(0,-1.0)/sqrt(2.0), std::complex<double>(-0.5,-0.5)/sqrt(2.0), std::complex<double>(1.0,-1.0)/sqrt(2.0), std::complex<double>(0,1.0)/sqrt(2.0)} },  
                                                                                        { {((-1.0 + sqrt(5))*(std::complex<double>(0,-1.0)*(3.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5)))))/(4.0*sqrt(2.0*(5.0 + sqrt(5)))), (-5.0 - sqrt(5))/10., 1.0, ((-1.0 + sqrt(5))*(std::complex<double>(0,1.0)*(3.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5)))))/(4.0*sqrt(2.0*(5.0 + sqrt(5))))}, {0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/(2.0*(-5.0 + sqrt(5))), (1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/4.0, 0.5+ (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {(1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0, (2.0 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(4.0*(-5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/8., (1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0}, {(std::complex<double>(0,0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(4.0*(-5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/8., (std::complex<double>(0,-0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {-((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 + sqrt(5))/6., 1.0, -((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {-(std::complex<double>(0,3.0) + sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(-1.0 + sqrt(5)))/12., ((1.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(-1.0 + sqrt(5)))/(2.0*(-3.0 + sqrt(5))), -(std::complex<double>(0,-3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(15)), ((1.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(-1.0 + sqrt(5)))/12., ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(-1.0 + sqrt(5)))/(2.0*(-3.0 + sqrt(5))), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(15))}, {((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 + sqrt(5))/6., 1.0, ((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))} },  
                                                                                        { {std::complex<double>(0,-0.5), -0.75, 1.0, std::complex<double>(0,0.5)}, {0.5, std::complex<double>(0,0.75), std::complex<double>(0,1.0), 0.5}, {std::complex<double>(0,0.5), 0.75, -1.0, std::complex<double>(0,-0.5)}, {std::complex<double>(0,0.5), 0.25 + std::complex<double>(0,1.0)/sqrt(2.0), (-1.0 + std::complex<double>(0,2.0)*sqrt(2.0))/3.0, std::complex<double>(0,-0.5)} },  
    /*#25*/                                                                             { {1.0 - std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))), (5.0 - sqrt(5))/10., 1.0, 1.0 + std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5)))}, {1.0 - (std::complex<double>(0,1.0)*(3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(4.0*(-5.0 + sqrt(5))), std::complex<double>(0,0.125)*(std::complex<double>(0,4.0)*(2.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), 1.0 + (std::complex<double>(0,1.0)*(3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {(std::complex<double>(0,1.0)*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (-1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/(2.0*(-5.0 + sqrt(5))), std::complex<double>(0,-0.125)*(std::complex<double>(0,-2.0)*(1.0 + sqrt(5)) + 3.0*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), -1.0 - (std::complex<double>(0,1.0)*(3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {-1.0 - (std::complex<double>(0,1.0)*(-3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (-1.0 + sqrt(5) + std::complex<double>(0,2.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(2.0*(5.0 + sqrt(5))), (-1.0 + sqrt(5) - std::complex<double>(0,2.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/4.0, (std::complex<double>(0,1.0)*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {0, -1.0, 1.0, 0}, {0, std::complex<double>(-1.0,-1.0)/sqrt(2.0), std::complex<double>(1.0,-1.0)/sqrt(2.0), 0}, {0, -1.0, 1.0, 0}, {std::complex<double>(0,1.0)/sqrt(2.0), std::complex<double>(-0.5,-0.5), std::complex<double>(0.5,-0.5), std::complex<double>(0,-1.0)/sqrt(2.0)} },  
                                                                                        { {std::complex<double>(0,1.0)/sqrt(3.0), -0.6666666666666666, 1.0, std::complex<double>(0,-1.0)/sqrt(3.0)}, {(std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), (-1.0 + std::complex<double>(0,3.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/12., (1.0 + std::complex<double>(0,3.0)*sqrt(3.0) - sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/8., (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {-((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/12., std::complex<double>(0,-0.125)*(std::complex<double>(0,-5) + sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15)), -((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,-3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), std::complex<double>(0,0.16666666666666666)*(sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5)), (std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/4.0, (std::complex<double>(0,3.0) + sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))} },  
                                                                                        { {(sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(-1.0 + sqrt(5)))/(4.0*sqrt(10 - 2.0*sqrt(5))), (-5.0 + sqrt(5))/10., 1.0, (1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/10.) - std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0))/4.0}, {0, (-5.0 + 3.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), (5.0 - 3.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-3.0 + sqrt(5))), 0}, {0, (std::complex<double>(0,-1.0)*sqrt((5.0 - sqrt(5))/2.0)*(-1.0 + sqrt(5)))/(-5.0 + sqrt(5)), (std::complex<double>(0,-1.0)*sqrt((5.0 - sqrt(5))/2.0)*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), 0}, {std::complex<double>(0,-1.0)/sqrt(10 - 2.0*sqrt(5)) + (-1.0 + sqrt(5))/4.0, (-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), -(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-3.0 + sqrt(5))), std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5)) + (-1.0 + sqrt(5))/4.0} },  
                                                                                        { {(std::complex<double>(0,-0.5)*(1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,0.5)*(1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), (3.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-3.0 + sqrt(5))), 0.5+ (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (4.0 - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (-4.0 - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), 0.5+ (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5))), -(1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(5.0 + sqrt(5))), (2.0 - 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) - std::complex<double>(0,3.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), ((1.0 + sqrt(5))*(sqrt(10 - 2.0*sqrt(5)) - std::complex<double>(0,1.0)*(-3.0 + sqrt(5))))/(4.0*sqrt(10 - 2.0*sqrt(5)))} },  
                                                                                        { {std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0)), -0.6666666666666666, 1.0, std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0))}, {0, std::complex<double>(0,1.0)*sqrt(0.6666666666666666), std::complex<double>(0,1.0)*sqrt(1.5), 0}, {0, 1.0/sqrt(2.0) - std::complex<double>(0,1.0)/sqrt(6), (std::complex<double>(0,-0.5)*(std::complex<double>(0,-3.0) + sqrt(3.0)))/sqrt(2.0), 0}, {std::complex<double>(0,1.0)*sqrt(0.6666666666666666), -sqrt(2.0)/3.0, 1.0/sqrt(2.0), std::complex<double>(0,-1.0)*sqrt(0.6666666666666666)} },  
                                                                                        { {(std::complex<double>(0,0.5)*(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (-5.0 - sqrt(5))/10., 1.0, -0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {0.5+ (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,-0.25)*(std::complex<double>(0,-2.0)*(1.0 + sqrt(5)) + 3.0*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(5.0 + sqrt(5)), (2.0 + 2.0*sqrt(5) - std::complex<double>(0,3.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), 0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0, (std::complex<double>(0,0.25)*(std::complex<double>(0,4.0)*(2.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(5.0 + sqrt(5)), (8 + 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), (-1.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0}, {-0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,0.25)*(std::complex<double>(0,4.0)*(2.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(5.0 + sqrt(5)), (8 + 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {(std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {(std::complex<double>(0,0.5)*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-4.0 - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-5.0 + sqrt(5))), (4.0 - std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(-3.0 + sqrt(5))), -0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {(std::complex<double>(0,0.5)*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (3.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-5.0 + sqrt(5))), (-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(2.0*(-3.0 + sqrt(5))), -0.5- (std::complex<double>(0,0.5)*(-3.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {std::complex<double>(0,0.25)*(std::complex<double>(0,1.0)*(1.0 + sqrt(5)) + sqrt((5.0 + sqrt(5))/2.0) - sqrt((5*(5.0 + sqrt(5)))/2.0)), 0, 0, std::complex<double>(0,0.125)*(std::complex<double>(0,2.0)*(1.0 + sqrt(5)) - sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5))))} },  
                                                                                        { {0.5+ (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (-5.0 - sqrt(5))/10., 1.0, 0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {(1.0 - sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0, (8 + 4.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), (std::complex<double>(0,0.25)*(std::complex<double>(0,4.0)*(2.0 + sqrt(5)) + sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(3.0 + sqrt(5)), (1.0 - sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10/(5.0 + sqrt(5))))/4.0}, {-0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) - std::complex<double>(0,3.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), (std::complex<double>(0,-0.25)*(std::complex<double>(0,-2.0)*(1.0 + sqrt(5)) + 3.0*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(3.0 + sqrt(5)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))}, {(std::complex<double>(0,-0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), 2.0/(-5.0 + sqrt(5)), 1.0, (std::complex<double>(0,0.25)*(-1.0 + sqrt(5))*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {(std::complex<double>(0,3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) - sqrt(15))/(4.0*sqrt(3.0)), -0.3333333333333333, 1.0, -(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {-((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-1.0 - sqrt(5))/6., (1.0 + sqrt(5))/2.0, -((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)), (std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {(std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)), (std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)} },  
                                                                                        { {0.5, -0.75, 1.0, 0.5}, {std::complex<double>(0,-0.25)*(std::complex<double>(1.0,-2.0) + sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(-3.0,-2.0) + sqrt(5)), (std::complex<double>(2.0,3.0) - std::complex<double>(0,1.0)*sqrt(5))/6., std::complex<double>(0,0.25)*(std::complex<double>(1.0,2.0) + sqrt(5))}, {(std::complex<double>(-1.0,-2.0) - sqrt(5))/4.0, (std::complex<double>(3.0,2.0) - sqrt(5))/8., (std::complex<double>(-3.0,2.0) + sqrt(5))/6., (std::complex<double>(-1.0,2.0) - sqrt(5))/4.0}, {(std::complex<double>(1.0,-2.0) + sqrt(5))/4.0, (std::complex<double>(-3.0,-2.0) + sqrt(5))/8., (std::complex<double>(3.0,-2.0) - sqrt(5))/6., (std::complex<double>(1.0,2.0) + sqrt(5))/4.0} },  
                                                                                        { {(std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), (-3.0 - sqrt(5))/6., 1.0, (std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), 0.3333333333333333, -2.0/(3.0 + sqrt(5)), ((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {-((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6., -((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/(3.0 + sqrt(5))), -((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {std::complex<double>(0,-1.0)/sqrt(3.0), (5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/12., -((5.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/(6 + 2.0*sqrt(5))), std::complex<double>(0,1.0)/sqrt(3.0)} },  
                                                                                        { {((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 + sqrt(5))/6., 1.0, ((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(3.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6., (5.0 + std::complex<double>(0,1.0)*sqrt(3.0) - sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/12., -(-5.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/(2.0*(-3.0 + sqrt(5))), (3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6.}, {(3.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6., std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,1.0) - 3.0*sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15)), (1.0 - std::complex<double>(0,3.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/(6 - 2.0*sqrt(5)), (3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6.}, {(3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6., (2.0 + std::complex<double>(0,1.0)*sqrt(3.0) - sqrt(5))/6., -((-2.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/(-3.0 + sqrt(5))), (3.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6.} },  
                                                                                        { {(3.0 + std::complex<double>(0,1.0)*sqrt(15))/6., -0.3333333333333333, 1.0, (3.0 - std::complex<double>(0,1.0)*sqrt(15))/6.}, {(3.0 + std::complex<double>(0,1.0)*sqrt(15))/6., (1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6., std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0)), (3.0 - std::complex<double>(0,1.0)*sqrt(15))/6.}, {(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), 0.3333333333333333, -1.0, (std::complex<double>(0,-3.0) - sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))}, {(1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/2.0, 0, 0, (1.0 - std::complex<double>(0,1.0)*sqrt(3.0))/2.0} },  
                                                                                        { {(std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(3.0), (-3.0 + sqrt(5))/6., 1.0, (std::complex<double>(0,-0.5)*(1.0 + sqrt(5)))/sqrt(3.0)}, {((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), -0.3333333333333333, (3.0 + sqrt(5))/2.0, ((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,1.0) + sqrt(3.0)), (1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/(3.0 - sqrt(5)), (std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0)), (5.0 - std::complex<double>(0,1.0)*sqrt(3.0) - sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/12., -(-5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/(2.0*(-3.0 + sqrt(5))), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0))} },  
                                                                                        { {(std::complex<double>(0,-0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) - std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5)), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,0.25)*(1.0 + sqrt(5))*(-3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5))))/sqrt(10 - 2.0*sqrt(5))}, {0.25 - sqrt(5)/4.0 + std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5)), (6 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,1.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(5.0 + sqrt(5))), std::complex<double>(0,0.125)*(sqrt(50 - 10*sqrt(5)) + sqrt(10 - 2.0*sqrt(5)) + std::complex<double>(0,2.0)*(3.0 + sqrt(5))), 0.25 - sqrt(5)/4.0 - std::complex<double>(0,1.0)/sqrt(10 - 2.0*sqrt(5))}, {(std::complex<double>(0,-1.0)*(-1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (std::complex<double>(0,-0.25)*(std::complex<double>(0,-4.0) + sqrt(50 - 10*sqrt(5)) + 3.0*sqrt(10 - 2.0*sqrt(5))))/(5.0 + sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(0,4.0) + sqrt(50 - 10*sqrt(5)) + 3.0*sqrt(10 - 2.0*sqrt(5))), (std::complex<double>(0,1.0)*(-1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))}, {(std::complex<double>(0,1.0)*(-1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5)), (4.0 + std::complex<double>(0,1.0)*sqrt(50 - 10*sqrt(5)) + std::complex<double>(0,3.0)*sqrt(10 - 2.0*sqrt(5)))/(4.0*(5.0 + sqrt(5))), std::complex<double>(0,0.125)*(std::complex<double>(0,4.0) + sqrt(50 - 10*sqrt(5)) + 3.0*sqrt(10 - 2.0*sqrt(5))), (std::complex<double>(0,-1.0)*(-1.0 + sqrt(5)))/sqrt(10 - 2.0*sqrt(5))} },  
                                                                                        { {(std::complex<double>(2.0,1.0) - std::complex<double>(0,1.0)*sqrt(5))/4.0, (-3.0 - sqrt(5))/8., 1.0, std::complex<double>(0,0.25)*(std::complex<double>(-1.0,-2.0) + sqrt(5))}, {std::complex<double>(0,-0.25)*(1.0 + sqrt(5)), std::complex<double>(0,0.125)*(std::complex<double>(1.0,2.0) + sqrt(5)), (std::complex<double>(2.0,1.0) + std::complex<double>(0,1.0)*sqrt(5))/(3.0 + sqrt(5)), std::complex<double>(0,0.25)*(1.0 + sqrt(5))}, {(-1.0 - sqrt(5))/4.0, (std::complex<double>(1.0,2.0) + sqrt(5))/8., -((std::complex<double>(1.0,-2.0) + sqrt(5))/(3.0 + sqrt(5))), (-1.0 - sqrt(5))/4.0}, {std::complex<double>(0,0.25)*(1.0 + sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(1.0,-2.0) + sqrt(5)), (std::complex<double>(0,-1.0)*(std::complex<double>(1.0,2.0) + sqrt(5)))/(3.0 + sqrt(5)), std::complex<double>(0,-0.25)*(1.0 + sqrt(5))} },  
                                                                                        { {std::complex<double>(0.5,0.5), -0.5, 1.0, std::complex<double>(0.5,-0.5)}, {std::complex<double>(-0.5,0.5), std::complex<double>(0,-0.5), std::complex<double>(0,-1.0), std::complex<double>(-0.5,-0.5)}, {std::complex<double>(0.5,-0.5), -0.5, 1.0, std::complex<double>(0.5,0.5)}, {(std::complex<double>(2.0,1.0) + std::complex<double>(0,1.0)*sqrt(5))/4.0, std::complex<double>(-0.125,0.125)*(-1.0 + sqrt(5)), std::complex<double>(0.25,0.25)*(-1.0 + sqrt(5)), std::complex<double>(0,-0.25)*(std::complex<double>(1.0,2.0) + sqrt(5))} },  
                                                                                        { {(std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), (-3.0 - sqrt(5))/6., 1.0, (std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {0, std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-3.0) + sqrt(3.0))*(1.0 + sqrt(5)), ((3.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/(2.0*(3.0 + sqrt(5))), 0}, {0, (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(3.0), (std::complex<double>(0,1.0)*sqrt(3.0)*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), 0}, {(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/12., (std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), (std::complex<double>(0,-3.0) - sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0))} },  
                                                                                        { {std::complex<double>(0,-0.25)*(-1.0 + sqrt(5)), (-5.0 - sqrt(5))/8., 1.0, std::complex<double>(0,0.25)*(-1.0 + sqrt(5))}, {std::complex<double>(0,0.25)*(1.0 + sqrt(5)), std::complex<double>(-0.25,0.5), std::complex<double>(2.0,4.0)/(5.0 + sqrt(5)), std::complex<double>(0,-0.25)*(1.0 + sqrt(5))}, {(1.0 + sqrt(5))/4.0, std::complex<double>(0.5,0.25), std::complex<double>(-4.0,2.0)/(5.0 + sqrt(5)), (1.0 + sqrt(5))/4.0}, {std::complex<double>(0,0.5), (std::complex<double>(-1.0,2.0) + std::complex<double>(1.0,2.0)*sqrt(5))/8., (std::complex<double>(1.0,2.0) - std::complex<double>(1.0,-2.0)*sqrt(5))/(5.0 + sqrt(5)), std::complex<double>(0,-0.5)} },  
                                                                                        { {-((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 - sqrt(5))/6., 1.0, -((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {0, (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(3.0), (std::complex<double>(0,1.0)*sqrt(3.0)*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), 0}, {0, ((3.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/12., (std::complex<double>(0,0.5)*(std::complex<double>(0,3.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), 0}, {0, std::complex<double>(0,0.16666666666666666)*(sqrt(3.0) + sqrt(15)), (std::complex<double>(0,1.0)*sqrt(3.0)*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), 0} },  
                                                                                        { {(std::complex<double>(0,-0.5)*(1.0 + sqrt(5)))/sqrt(3.0), (-3.0 + sqrt(5))/6., 1.0, (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(3.0)}, {std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,3.0) + sqrt(3.0)), std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-1.0) - 3.0*sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15)), (1.0 + std::complex<double>(0,3.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/(6 - 2.0*sqrt(5)), std::complex<double>(0,-0.16666666666666666)*(std::complex<double>(0,-3.0) + sqrt(3.0))}, {std::complex<double>(0,-1.0)/sqrt(3.0), (5.0 - std::complex<double>(0,1.0)*sqrt(3.0) - sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/12., -(-5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/(2.0*(-3.0 + sqrt(5))), std::complex<double>(0,1.0)/sqrt(3.0)}, {-(std::complex<double>(0,3.0) - sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) + sqrt(15))/(4.0*sqrt(3.0)), std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)), (std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(-3.0 + sqrt(5)), (std::complex<double>(0,3.0) + sqrt(3.0) + std::complex<double>(0,1.0)*sqrt(5) - sqrt(15))/(4.0*sqrt(3.0))} },  
                                                                                        { {-((std::complex<double>(0,1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0)), (-3.0 + sqrt(5))/6., 1.0, -((std::complex<double>(0,-1.0) + sqrt(3.0))*(1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(3.0 + std::complex<double>(0,1.0)*sqrt(3.0))/6., (-5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) + std::complex<double>(0,1.0)*sqrt(15))/12., (-5.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/(2.0*(-3.0 + sqrt(5))), (3.0 - std::complex<double>(0,1.0)*sqrt(3.0))/6.}, {std::complex<double>(0,1.0)/sqrt(3.0), (-2.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/6., (-2.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/(-3.0 + sqrt(5)), std::complex<double>(0,-1.0)/sqrt(3.0)}, {std::complex<double>(0,1.0)/sqrt(3.0), (-2.0 + std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/6., (-2.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5))/(-3.0 + sqrt(5)), std::complex<double>(0,-1.0)/sqrt(3.0)} },  
                                                                                        { {std::complex<double>(0,-1.0)*sqrt(2.0/(5.0 + sqrt(5))), (-5.0 - sqrt(5))/10., 1.0, std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5)))}, {0, (std::complex<double>(0,1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,1.0)*(1.0 + sqrt(5))*sqrt((5.0 + sqrt(5))/2.0))/(3.0 + sqrt(5)), 0}, {0, (5.0 + 3.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/(10 + 2.0*sqrt(5)), -((5.0 + 3.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))))/(6 + 2.0*sqrt(5))), 0}, {-0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,0.25)*(std::complex<double>(0,2.0)*(1.0 + sqrt(5)) + 3.0*sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))))/(5.0 + sqrt(5)), (2.0 + 2.0*sqrt(5) + std::complex<double>(0,3.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(12 + 4.0*sqrt(5)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {(std::complex<double>(0,-1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (-5.0 + sqrt(5))/10., 1.0, (std::complex<double>(0,1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {0, std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))), std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0), 0}, {0, (10 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(0,-2.0)*(5.0 + sqrt(5)) - sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), 0}, {-0.5- (std::complex<double>(0,0.5)*(3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), -(2.0 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(4.0*(5.0 + sqrt(5))), (2.0 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/8., (std::complex<double>(0,0.5)*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {(std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), (-3.0 - sqrt(5))/6., 1.0, (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {(std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), ((1.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(3.0 + sqrt(5)))/12., std::complex<double>(0,0.5)*(std::complex<double>(0,1.0) + sqrt(3.0)), (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {-((std::complex<double>(0,-1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0)), ((1.0 - std::complex<double>(0,1.0)*sqrt(3.0))*(3.0 + sqrt(5)))/12., std::complex<double>(0,-0.5)*(std::complex<double>(0,-1.0) + sqrt(3.0)), -((std::complex<double>(0,1.0) + sqrt(3.0))*(-1.0 + sqrt(5)))/(4.0*sqrt(3.0))}, {(std::complex<double>(0,-0.5)*(1.0 + sqrt(5)))/sqrt(3.0), std::complex<double>(0,0.16666666666666666)*(std::complex<double>(0,1.0) + sqrt(3.0)), (1.0 + std::complex<double>(0,1.0)*sqrt(3.0))/(3.0 + sqrt(5)), (std::complex<double>(0,0.5)*(1.0 + sqrt(5)))/sqrt(3.0)} },  
                                                                                        { {0, -1.0, 1.0, 0}, {(std::complex<double>(0,1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (10 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), -0.5- (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,-1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {(1.0 + sqrt(5))/4.0 + std::complex<double>(0,1.0)/sqrt(2.0*(5.0 + sqrt(5))), std::complex<double>(0,-1.0)*sqrt(2.0/(5.0 + sqrt(5))), std::complex<double>(0,-1.0)*sqrt(2.0/(5.0 + sqrt(5))), (1.0 + sqrt(5))/4.0 - std::complex<double>(0,1.0)/sqrt(2.0*(5.0 + sqrt(5)))}, {(std::complex<double>(0,-1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (10 + 2.0*sqrt(5) - std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) + std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), -0.5+ (std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5))), (std::complex<double>(0,1.0)*(1.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))} },  
                                                                                        { {std::complex<double>(0,0.5), -0.75, 1.0, std::complex<double>(0,-0.5)}, {0.5, std::complex<double>(0,-0.25)*(std::complex<double>(0,-2.0) + sqrt(5)), (2.0 - std::complex<double>(0,1.0)*sqrt(5))/3.0, 0.5}, {std::complex<double>(0,0.5), (std::complex<double>(0,2.0) - sqrt(5))/4.0, (std::complex<double>(0,2.0) + sqrt(5))/3.0, std::complex<double>(0,-0.5)}, {std::complex<double>(0,-0.5), (std::complex<double>(0,-2.0) + sqrt(5))/4.0, (std::complex<double>(0,-2.0) - sqrt(5))/3.0, std::complex<double>(0,0.5)} },  
                                                                                        { {(std::complex<double>(0,0.5)*(3.0 + sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5)))))/sqrt(2.0*(5.0 + sqrt(5))), (-5.0 + sqrt(5))/10., 1.0, -0.5- (std::complex<double>(0,0.5)*(3.0 + sqrt(5)))/sqrt(2.0*(5.0 + sqrt(5)))}, {0, std::complex<double>(0,1.0)*sqrt(2.0/(5.0 + sqrt(5))), std::complex<double>(0,1.0)*sqrt((5.0 + sqrt(5))/2.0), 0}, {0, (10 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(0,-2.0)*(5.0 + sqrt(5)) - sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), 0}, {0, (10 + 2.0*sqrt(5) + std::complex<double>(0,1.0)*sqrt(2.0*(5.0 + sqrt(5))) - std::complex<double>(0,1.0)*sqrt(10*(5.0 + sqrt(5))))/(20 + 4.0*sqrt(5)), std::complex<double>(0,-0.125)*(std::complex<double>(0,-2.0)*(5.0 + sqrt(5)) - sqrt(2.0*(5.0 + sqrt(5))) + sqrt(10*(5.0 + sqrt(5)))), 0} },  
    /*#54*/                                                                              { {(std::complex<double>(0,0.5)*(-1.0 + sqrt(5)))/sqrt(3.0), (-3.0 - sqrt(5))/6., 1.0, (std::complex<double>(0,-0.5)*(-1.0 + sqrt(5)))/sqrt(3.0)}, {0, std::complex<double>(0,0.08333333333333333)*(std::complex<double>(0,3.0) + sqrt(3.0))*(1.0 + sqrt(5)), ((3.0 + std::complex<double>(0,1.0)*sqrt(3.0))*(1.0 + sqrt(5)))/(2.0*(3.0 + sqrt(5))), 0}, {0, std::complex<double>(0,0.16666666666666666)*(sqrt(3.0) + sqrt(15)), (std::complex<double>(0,1.0)*sqrt(3.0)*(1.0 + sqrt(5)))/(3.0 + sqrt(5)), 0}, {std::complex<double>(0,1.0)/sqrt(3.0), std::complex<double>(0,-0.08333333333333333)*(std::complex<double>(0,-5) + sqrt(3.0) - std::complex<double>(0,1.0)*sqrt(5) + sqrt(15)), (5.0 - std::complex<double>(0,1.0)*sqrt(3.0) + sqrt(5) - std::complex<double>(0,1.0)*sqrt(15))/(6 + 2.0*sqrt(5)), std::complex<double>(0,-1.0)/sqrt(3.0)} } };  
   
    int max_L = 200;
    double my_small = 0.0000001;
    std::vector< std::vector< std::complex<double> > > orbit;
    for(int kk = 0; kk < classification.size(); kk++){
        std::cout << "Orbit #" << (kk+1) << std::endl; 
        std::complex<double> m1_11 = classification.at(kk).at(0).at(0);
        std::complex<double> m1_12 = classification.at(kk).at(0).at(1);
        std::complex<double> m1_21 = classification.at(kk).at(0).at(2);
        std::complex<double> m1_22 = classification.at(kk).at(0).at(3);
        std::complex<double> m2_11 = classification.at(kk).at(1).at(0);
        std::complex<double> m2_12 = classification.at(kk).at(1).at(1);
        std::complex<double> m2_21 = classification.at(kk).at(1).at(2);
        std::complex<double> m2_22 = classification.at(kk).at(1).at(3);
        std::complex<double> m3_11 = classification.at(kk).at(2).at(0);
        std::complex<double> m3_12 = classification.at(kk).at(2).at(1);
        std::complex<double> m3_21 = classification.at(kk).at(2).at(2);
        std::complex<double> m3_22 = classification.at(kk).at(2).at(3);
        std::complex<double> m4_11 = classification.at(kk).at(3).at(0);
        std::complex<double> m4_12 = classification.at(kk).at(3).at(1);
        std::complex<double> m4_21 = classification.at(kk).at(3).at(2);
        std::complex<double> m4_22 = classification.at(kk).at(3).at(3);    
        std::vector< std::complex<double> > m1 = {m1_11,m1_12,m1_21,m1_22};
        std::vector< std::complex<double> > m2 = {m2_11,m2_12,m2_21,m2_22};
        std::vector< std::complex<double> > m3 = {m3_11,m3_12,m3_21,m3_22};
        std::vector< std::complex<double> > m4 = {m4_11,m4_12,m4_21,m4_22};
        orbit.clear();
        if(!check_is_in(m1,orbit)){
            orbit.push_back(m1);
        }
        if(!check_is_in(m2,orbit)){
            orbit.push_back(m2);
        }
        if(!check_is_in(m3,orbit)){
            orbit.push_back(m3);
        }
        if(!check_is_in(m4,orbit)){
            orbit.push_back(m4);
        }
        int offset = 0;
        int old_orbit_size;
       
        do{ 
            old_orbit_size = orbit.size();
            for(int ii = (0 + offset); ii < old_orbit_size; ii++){
                std::vector< std::complex<double> > m5 = matrix_multiplication(orbit.at(ii),m1);
                if(!check_is_in(m5,orbit)){
                    orbit.push_back(m5);
                }
                std::vector< std::complex<double> > m6 = matrix_multiplication(orbit.at(ii),m2);
                if(!check_is_in(m6,orbit)){
                    orbit.push_back(m6);
                }
                std::vector< std::complex<double> > m7 = matrix_multiplication(orbit.at(ii),m3);
                if(!check_is_in(m7,orbit)){
                    orbit.push_back(m7);
                }
                std::vector< std::complex<double> > m8 = matrix_multiplication(orbit.at(ii),m4);
                if(!check_is_in(m8,orbit)){
                    orbit.push_back(m8);
                }
            }   
            if( (old_orbit_size - orbit.size()) == 0 ){
                break;
            };
            offset = old_orbit_size;
        }while( orbit.size() < max_L );
        if( orbit.size() > max_L ){
            std::cout << "Err: The group <M1,...,M4> has order bigger than " << 120 << "." << std::endl;
        }else{
            std::cout << "The group <M1,...,M4> has order: " << orbit.size() << std::endl;
        }
        std::cout << "------" << std::endl;  
        //look for duplicates:
        std::vector< std::vector< std::complex<double> > > orbit_tmp = {};
    }

    return 0; 

}